# -*- encoding: utf-8 -*-

from django.utils.functional import curry
from django.forms import widgets
from django.db import models

NOM_CHAMPS_ETAT = 'etat'
TEXTE_CHAMPS_ETAT = 'Statut (étape)'
HELP_CHAMPS_ETAT = "Ce champs montre par défaut l'étape où est rendu le contenu. La liste déroulante présente les changements d'états autorisés."

class WorkflowMixin(models.Model):
    """
    Classe abstraite pour implanter un workflow simple.
    """
    # conserve le request pour connaitre l'utilisateur connecté, et le rendre
    # disponible dans les fonctions de d'accès et d'avertissement
    _request = None

    # conserve une trace de l'état courant car la variable 'etat' prend la nouvelle
    # valeur après validation mais on veut conserver cette information pour faire des
    # fonctions d'avertissements plus précise en fonction de l'action
    # (caractérisée par un état initial et un état final).
    _etat_courant = None

    # variable utilisées pour définir le workflow
    etat_initial = None
    etats = {}
    actions = {}

    # nouveau champ ajouté dans la classe fille
    etat = models.CharField(verbose_name=TEXTE_CHAMPS_ETAT, help_text=HELP_CHAMPS_ETAT, max_length=20, default=etat_initial, blank = True, null = True)

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super(WorkflowMixin, self).__init__(*args, **kwargs)
        
        field_object, model, direct, m2m = self._meta.get_field_by_name(NOM_CHAMPS_ETAT)

        # test si les états définis respectent les critères de validations du champs 'état'
        for e in self.etats.keys():
            if len(e) >= field_object.max_length:
                raise Exception(u"Nom de l'état trop long :'%s' (max: %s)" % (e, field_object.max_length))

        # genère l'accesseur pour l'objet
        setattr(self.__class__, 'get_%s_display' % NOM_CHAMPS_ETAT, curry(self.__class__._get_FIELD_display, field=field_object))

        field_object.default = self.etat_initial
        # fournir tous les états disponibles pour que la validation soit acceptée.
        # le filtrage s'opère dans le rendu du form dans l'admin du Worflow
        if self.id:
            field_object._choices = [(k, v) for k,v in self.etats.items()]
        # dans le cas d'un nouvel objet, limiter la liste à son état initial
        else:
            field_object._choices = [(k, v) for k,v in self.etats.items() if k == self.etat_initial]
        field_object.widget = widgets.Select()

        # initialisation de l'état par l'état initial prédéfini
        if self.etat is None:
            self.etat = self.etat_initial

    def test_droit_action(self, code_action, action, request):
        """
        Fonction utilisée pour savoir si une action est disponible.
        Elle s'assure tout d'abord que l'action testée possède un état
        initial identique à l'état courant de l'objet.
        Ensuite, elle appelle une fonction respectant une convention de nommage,
        pour savoir si l'action est disponible.
        """
        
        # recherche et appel de la fonction qui déterminera si l'action est 
        # véritablement disponible, autrement une exception est propagée
        nom_fonction_droit = "acces_%s" % code_action
        nom_fonction_droit_lower = nom_fonction_droit.lower()
        fonction_droit = getattr(self, nom_fonction_droit_lower, None)
        if fonction_droit:
            return fonction_droit(action, request)
        else:
            raise Exception("Méthode manquante dans la classe de Workflow : '%s'" % nom_fonction_droit_lower)

    def get_action(self, etat_initial, etat_final):
        """
        Rechercher l'action en fonction des états.
        Si les états sont identiques, None est retourné (pas d'action).
        """
        # pas d'action sans changement d'état
        if etat_initial == etat_final:
            return None

        for action, proprietes in self.actions.items():
            if getattr(proprietes['etat_initial'], '__iter__', False):
                match_initial = etat_initial in proprietes['etat_initial']
            else:
                match_initial = etat_initial == proprietes['etat_initial']

            if match_initial and  proprietes['etat_final'] == etat_final:
               return action
        raise Exception("Ces 2 états ne permettent pas d'identifier une action (initial:%s final:%s)" % (etat_initial, etat_final))
    
    def get_etats_disponibles(self, request=None):
        """
        Retourne la liste des états suivants possibles, en tenant compte des permissions.
        """
        self._request = request
        prochains_etats = [proprietes['etat_final'] for action, proprietes in self.actions.items() if self.test_droit_action(action, proprietes, request)]
    
        if not self.etats.has_key(self.etat):
            aide = """Si la valeur non permise est 'None', vous avez peut-être introduit le Workflow après que du contenu était déjà présent.
            Dans ce cas, vous devez migrer ces anciens contenus en leur spécifiant une étape de workflow valide"""
            raise Exception("valeur non permise %s (%s)" % (self.etat, aide))

        # en premier, garder l'état courant
        etats = [(self.etat,  self.etats[self.etat]), ]

        #puis cumuler les états suivantes disponible à partir de l'état courant
        for e in prochains_etats:
            etats.append((e, self.etats[e]))
        return etats

    def __setattr__(self, name, value):
        """
        Contrôle de l'assignation des variables, et traitement spécial pour l'état
        afin qu'il puisse être conservé.
        """
        if name == NOM_CHAMPS_ETAT and self.__dict__.has_key(name) and self.__dict__[name] != value:
            self._etat_courant = self.__dict__[name]
            action_valide = self.get_action(self._etat_courant, value)
        return object.__setattr__(self, name, value)


    def save(self, force_insert=False, force_update=False, using=None):
        """
        Surcharge de la sauvegarde pour éventuellement exécuter la fonction de changement d'état.
        """
        # s'assurer que l'état est valide
        # (devrait être testé au moment de l'assignation)

        super(WorkflowMixin, self).save(force_insert, force_update, using)

        # execution de traitement spécial de changement d'état (optionnel)
        nom_fonction_transition = "passage_de_%s_a_%s" % (self._etat_courant , self.etat)
        nom_fonction_transition_lower = nom_fonction_transition.lower()

        
        try:
            fonction_transition = getattr(self, nom_fonction_transition_lower, None)
        except:
            raise Exception(u"le nom généré de la fonction de transition est incorrect : %s" % fonction_transition)
            
        if fonction_transition:
            fonction_transition(self.get_action(self._etat_courant, self.etat), self._request)
        
        # sauvegarde l'état courant
        self._etat_courant = self.etat

